import os
import stat
import tempfile

from zope import interface

from zope.app.file.interfaces import IImage

LengthError = (TypeError, AttributeError)


class VImage(object):

    """a non persistent image implementation"""

    interface.implements(IImage)

    def __init__(self, contentType='', size=(0, 0)):
        self.data = tempfile.TemporaryFile(prefix='z3c.vimage')
        self.contentType = contentType
        self.size = size
                 
    def getSize(self):
        try:
            return len(self.data)
        except LengthError:
            data = self.data
            if hasattr(data, 'fileno'):
                return int(os.fstat(data.fileno())[stat.ST_SIZE])
        pos = self.data.tell()
        size = len(data.read())
        data.seek(pos)
        return size

    def getImageSize(self):
        return self.size
    



import time
import PIL
from datetime import datetime
from cStringIO import StringIO
from types import StringType

import zope.datetime

from zope.security.proxy import isinstance, removeSecurityProxy
from zope.publisher.browser import BrowserView
from zope.dublincore.interfaces import IZopeDublinCore

from zope.app.file.image import getImageInfo

from interfaces import IProcessableImage


def getMaxSize(image_size, desired_size):
    """returns the maximum size of image_size to fit into the
    rectangualar defined by desired_size

    >>> getMaxSize((100,200),(100,100))
    (50, 100)

    >>> getMaxSize((100,300),(100,100))
    (33, 100)

    >>> getMaxSize((300,100),(100,100))
    (100, 33)

    >>> getMaxSize((300,100),(400,200))
    (400, 133)


    """
    x_ratio = float(desired_size[0])/image_size[0]
    y_ratio = float(desired_size[1])/image_size[1]
    if x_ratio < y_ratio:
        new_size = (round(x_ratio*image_size[0]),
                    round(x_ratio*image_size[1]))
    else:
        new_size = (round(y_ratio*image_size[0]),
                    round(y_ratio*image_size[1]))
    return tuple(map(int, new_size))


class ImageProcessorView(BrowserView):

    """image processor"""

    def __init__(self, context, request):
        self.params = request.form
        super(ImageProcessorView,self).__init__(context,request)

    def _collectParams(self):
        # rotate
        self.degrees =int(self.params.get('remote.adjust.rotate',0))
        # resize
        self.width = int(self.params.get('remote.size.w',0))
        self.height = int(self.params.get('remote.size.h',0))
        # crop
        self.cropX = self.params.get('local.crop.x',None)
        self.cropW = self.params.get('local.crop.w',None)
        self.cropY = self.params.get('local.crop.y',None)
        self.cropH = self.params.get('local.crop.h',None)
        # rezize after crop
        self.afterSizeW = int(self.params.get('after.size.w',0))
        self.afterSizeH = int(self.params.get('after.size.h',0))
        self.size = (self.width,self.height)
        self._calcAfterSize()

    @property
    def data(self):
        context = removeSecurityProxy(self.context)
        data = context.data
        if not isinstance(data, str):
            pos = data.tell()
            data.seek(0)
            res = data.read(1024*1024)
            data.seek(pos)
            data = res
        else:
            data = context.data
        return data

    def _resultingRatio(self):
        if (    self.cropW is not None
            and self.cropH is not None
            and self.cropW != 0
            and self.cropH != 0
           ):
            ratio = float(self.cropW) / float(self.cropH)
        else:
            t,w,h = getImageInfo(self.data)
            ratio = float(w) / float(h)
        return ratio

    def _calcAfterSize(self):
        if (    (self.afterSizeW == 0 or self.afterSizeH == 0)
            and self.afterSizeW != self.afterSizeH
           ):
            ratio = self._resultingRatio()
            if self.afterSizeH == 0:
                self.afterSizeH = int(round(self.afterSizeW / ratio))
            if self.afterSizeW == 0:
                self.afterSizeW = int(round(self.afterSizeH * ratio))
        self.afterSize = (self.afterSizeW,self.afterSizeH)

    def _process(self):
        self._collectParams()
        pimg = IProcessableImage(self.context)
        self._pushCommands(pimg)
        return pimg.process()

    def _pushCommands(self, pimg):
        if self.degrees > 0:
            pimg.rotate(self.degrees)
        if self.width and self.height:
            pimg.resize(self.size)
        if self.cropX is not None and self.cropY is not None \
           and self.cropW is not None and self.cropH is not None:
            self.croparea = (int(self.cropX),
                             int(self.cropY),
                             int(self.cropX) + int(self.cropW),
                             int(self.cropY) + int(self.cropH))
            pimg.crop(self.croparea)
        if self.afterSizeW is not 0 and self.afterSizeH is not 0:
            if self.cropW and self.cropH:
                currentSize = (int(self.cropW), int(self.cropH))
            elif self.width and self.height:
                currentSize = (self.width, self.height)
            else:
                t,w,h = getImageInfo(self.data)
                currentSize = (w,h)
            size=getMaxSize(currentSize, self.afterSize)
            pimg.resize(size)

    def processed(self):
        processed = self._process()
        if self.request is not None:
            self.request.response.setHeader('Content-Type',
                                            processed.contentType)
            self.request.response.setHeader('Content-Length',
                                            processed.getSize())
        return processed.data

    def __call__(self):
        try:
            modified = IZopeDublinCore(self.context).modified
        except TypeError:
            modified=None
        if modified is None or not isinstance(modified,datetime):
            return self.processed()
        header= self.request.getHeader('If-Modified-Since', None)
        lmt = long(time.mktime(modified.timetuple()))
        if header is not None:
            header = header.split(';')[0]
            try:    mod_since=long(zope.datetime.time(header))
            except: mod_since=None
            if mod_since is not None:
                if lmt <= mod_since:
                    self.request.response.setStatus(304)
                    return ''
        self.request.response.setHeader('Last-Modified',
                                        zope.datetime.rfc1123_date(lmt))
        return self.processed()


class ResizedImageView(ImageProcessorView):

    def _collectParams(self):
        super(ResizedImageView,self)._collectParams()
        data = self.data
        t,w,h = getImageInfo(data)
        self.size = (w,h)
        self.width = self.params.get('w',self.size[0])
        self.height = self.params.get('h',self.size[1])

    def _pushCommands(self, pimg):
        new_size = getMaxSize(self.size, (self.width, self.height))
        if new_size != self.size:
            pimg.resize(new_size)


class PasteImageView(ResizedImageView):
    """Paste an image into the image.

    The subclass must provide the image to be pasted.
    """

    img = None

    def _collectParams(self):
        super(PasteImageView, self)._collectParams()
        self.x = self.params.get('x', 0)
        self.y = self.params.get('y', 0)
        self.imgId = self.params.get('img', 0)

    def _pushCommands(self, pimg):
        super(PasteImageView, self)._pushCommands(pimg)
        img = self.img
        if img is not None:
            if type(img)!=StringType:
                img.seek(0)
                img = img.read()
            img = PIL.Image.open(StringIO(img))
            pimg.paste(img, (self.x, self.y), img)




from zope.interface import implements
from zope import component
from z3c.image.image import VImage
from zope.app.file.interfaces import IFile
from zope.cachedescriptors.property import readproperty
from cStringIO import StringIO
from interfaces import IProcessableImage
from PIL import ImageFile, Image
from types import StringType
from zope.app.cache.ram import RAMCache
import os
import logging
try:
    ulimit = int(os.popen('ulimit -n').read().strip())
    maxEntries = ulimit - 100
except:
    ulimit = -1
    maxEntries = 100

# see http://mail.python.org/pipermail/image-sig/2003-May/002228.html
ImageFile.MAXBLOCK = 1024*1024*10
imgCache = RAMCache()
maxEntries = min(maxEntries, 400)
logging.debug('z3c.image.proc init imgCache maxEntries: %r, ulimit %r' % (
    maxEntries, ulimit))
imgCache.maxEntries = maxEntries

def invalidateCache(object, event):
    imgCache.invalidate(object)


class ProcessableImage(object):

    component.adapts(IFile)
    implements(IProcessableImage)

    def __init__(self, image):
        self.context = image
        self.format = ''
        contentType = image.contentType
        if contentType is not None:
            try:
                self.format = image.contentType.split('/')[1]
            except IndexError:
                pass
        self.cmds = []

    def getPILImg(self):
        data = self.context.data
        if type(data)!=StringType:
            data.seek(0)
            data = data.read()
        p = Image.open(StringIO(data))
        return p

    def _toImage(self, pimg, *args,**kw):
        """returns an Image object from the given PIL image"""
        if self.format == 'gif':
            # optimization seems to create corrupted gif files wiht
            # PIL, so we switch it off
            kw.pop('optimize')
        img = VImage(contentType=self.context.contentType,
                     size=pimg.size)
        try:
            pimg.save(img.data,self.format,*args,**kw)
        except IOError:
            # retry without optimization
            kw.pop('optimize')
            pimg.save(img.data,self.format,*args,**kw)
        img.data.seek(0)
        return img

    def rotate(self, degrees):
        self.cmds.append(('rotate',(degrees,),{}))

    def crop(self, croparea):
        croparea = map(int,croparea)
        self.cmds.append(('crop',(croparea,),{}))

    def resize(self, size):
        """See IPILImageResizeUtility"""
        size = map(int, size)
        self.cmds.append(('resize',(size, Image.ANTIALIAS),{}))

    def paste(self, img, box=None, mask=None):
        if box is not None:
            box = tuple(map(int, box))
        self.cmds.append(('paste',(img, box, mask), {}))

    def reset(self):
        self.cmds=[]

    def process(self,quality=90,optimize=1):
        """processes the command queue and returns the image"""
        if not self.cmds:
            return self.context
        key = {'cmds':str(self.cmds)}
        img = imgCache.query(self.context , key)
        if img is not None and not img.data.closed:
            img.data.seek(0)
            return img
        pimg = self.getPILImg()

        for name,args,kwords in self.cmds:
            func = getattr(pimg,name)
            oldImg = pimg
            pimg = func(*args,**kwords)
            if pimg is None:
                pimg = oldImg

        img = self._toImage(pimg, quality=quality, optimize=optimize)
        imgCache.set(img, self.context, key=key)
        return img



#testing package
import os
from zope.app.file.image import Image

dataDir = os.path.join(os.path.dirname(__file__),'data')

def readTestImage(name):
    path = os.path.join(dataDir,name)
    return file(path, 'rb').read()

def getTestImage(name):
    """returns a zope image with the given name from this directory"""
    return Image(readTestImage(name))




import doctest
import unittest
from zope.testing.doctestunit import DocFileSuite, DocTestSuite
from zope.app.testing import functional
from zope import component
import adapter
from zope.app.file.interfaces import IImage
from zope.lifecycleevent.interfaces import IObjectModifiedEvent


functional.defineLayer('Z3cImageLayer', 'ftesting.zcml')


def setUp(test):
    component.provideHandler(adapter.invalidateCache,
                             adapts=(IImage, IObjectModifiedEvent))

def tearDown(test):
    pass

def test_suite():
    ftests = functional.FunctionalDocFileSuite('BROWSER.txt')
    ftests.layer = Z3cImageLayer
    return unittest.TestSuite(
        (
        ftests,
        DocFileSuite('README.txt',
                     setUp=setUp,
                     optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
                     ),
        DocTestSuite('z3c.image.proc.browser',
                     optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
                     ),
        ))


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



from zope.interface import Interface

class IProcessableImage(Interface):

    def process():
        """returns the processed image"""

    def rotate(degrees):
        """rotates the image by degrees"""

    def resize(size):
        """resizes the image to (w,h)"""

    def crop(croparea):
        """ crops the image """

    def paste(pos, img, mask):
        """ paste an image at pos with mask"""




from setuptools import setup, find_packages

setup(
    name = "z3c.image",
    version = "1.0.0",
    author = "Zope Contributors",
    author_email = "office@mopacreative.com",
    description = "Image utils for zope3",
    license = "ZPL 2.1",
    keywords = "zope3 image",
    url='http://svn.zope.org/z3c.image',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        "License :: OSI Approved :: Zope Public License",
        "Framework :: Zope3",
        ],
    packages = find_packages('src'),
    include_package_data = True,
    package_dir = {'':'src'},
    namespace_packages = ['z3c'],
    extras_require = dict(test=['zope.app.testing',
                                'zope.app.server',
                                'zope.app.authentication',
                                'zope.app.securitypolicy',
                                'zope.app.zcmlfiles',
                                'zope.lifecycleevent',
                                'zope.testbrowser',
                                'zope.testing',
                               ]),
    install_requires = ['setuptools',
    # 'PILwoTk' is a package containing a version of PIL that doesn't
    # magically sniff for the Tk installation.
                        'PILwoTk',
                        'zope.app.cache',
                        'zope.app.file',
                        'zope.cachedescriptors',
                        'zope.datetime',
                        'zope.dublincore',
                        'zope.interface',
                        'zope.publisher',
                        'zope.security',
                        ],
    zip_safe=False,
    )
